@@ -1,5 +1,5 @@
 __author__ = "Shyue Ping Ong"
-__date__ = "Oct 16 2013"
-__version__ = "0.5.1"
+__date__ = "Dec 21 2013"
+__version__ = "0.6.0"
 
 from custodian import Custodian
