@@ -21,7 +21,7 @@ import subprocess
 import datetime
 import time
 import json
-import glob
+from glob import glob
 import tarfile
 import os
 import tempfile
@@ -29,6 +29,8 @@ import shutil
 from abc import ABCMeta, abstractmethod, abstractproperty
 from gzip import GzipFile
 
+pjoin = os.path.join
+
 
 class Custodian(object):
     """
@@ -67,15 +69,16 @@ class Custodian(object):
     .. attribute: monitor_freq
 
         The number of polling steps before monitoring occurs. For example,
-        if you have a polling_time_step of 10seconds and a monitor_freq of
+        if you have a polling_time_step of 10 seconds and a monitor_freq of
         30, this means that Custodian uses the monitors to check for errors
         every 30 x 10 = 300 seconds, i.e., 5 minutes.
     """
+    LOG_FILE = "custodian.json"
 
     def __init__(self, handlers, jobs, max_errors=1, polling_time_step=10,
                  monitor_freq=30, log_file="custodian.json",
                  skip_over_errors=False, scratch_dir=None,
-                 gzipped_output=False):
+                 gzipped_output=False, checkpoint=True):
         """
         Args:
             handlers:
@@ -94,8 +97,8 @@ class Custodian(object):
                 monitors to check for errors every 30 x 10 = 300 seconds,
                 i.e., 5 minutes.
             log_file:
-                Log file to log all jobs and corrections to. Defaults
-                to custodian.json. Set to None for no logging.
+                Deprecated. Custodian now always logs to a custodian.json
+                file.
             skip_over_errors:
                 If set to True, custodian will skip over error handlers that
                 failed (raised an Exception of some sort). Otherwise,
@@ -117,6 +120,8 @@ class Custodian(object):
             gzipped_output:
                 Whether to gzip the final output to save space. Defaults to
                 False.
+            checkpoint:
+                Whether to checkpoint after each Job. Defaults to True.
         """
         self.max_errors = max_errors
         self.jobs = jobs
@@ -124,10 +129,35 @@ class Custodian(object):
         self.monitors = filter(lambda x: x.is_monitor, handlers)
         self.polling_time_step = polling_time_step
         self.monitor_freq = monitor_freq
-        self.log_file = log_file
         self.skip_over_errors = skip_over_errors
         self.scratch_dir = scratch_dir
         self.gzipped_output = gzipped_output
+        self.checkpoint = checkpoint
+
+    @staticmethod
+    def _load_checkpoint(fname):
+        logging.info("Loading from checkpoint file {}...".format(fname))
+        t = tarfile.open(fname)
+        t.extractall()
+        #Log the corrections to a json file.
+        with open(Custodian.LOG_FILE, "r") as f:
+            run_log = json.load(f)
+        return run_log
+
+    @staticmethod
+    def _delete_checkpoints(cwd):
+        for f in glob(pjoin(cwd, "custodian.chk.*.tar.gz")):
+            os.remove(f)
+
+    @staticmethod
+    def _save_checkpoint(cwd, index):
+        try:
+            Custodian._delete_checkpoints(cwd)
+            name = shutil.make_archive(
+                pjoin(cwd, "custodian.chk.{}".format(index)), "gztar")
+            logging.info("Checkpoint written to {}".format(name))
+        except Exception as ex:
+            logging.info("Checkpointing failed")
 
     def run(self):
         """
@@ -137,38 +167,26 @@ class Custodian(object):
             All errors encountered as a list of list.
             [[error_dicts for job 1], [error_dicts for job 2], ....]
         """
+        cwd = os.getcwd()
+
         if self.scratch_dir is not None:
-            cwd = os.getcwd()
             tempdir = tempfile.mkdtemp(dir=self.scratch_dir)
             for f in os.listdir("."):
                 shutil.copy(f, tempdir)
             os.chdir(tempdir)
+            logging.info("Using scratch directory {}.".format(tempdir))
 
         run_log = []
         total_errors = 0
         unrecoverable = False
         start = datetime.datetime.now()
         logging.info("Run started at {}.".format(start))
-        def do_check(handlers, terminate_func=None):
-            corrections = []
-            for h in handlers:
-                try:
-                    if h.check():
-                        if terminate_func is not None:
-                            terminate_func()
-                        d = h.correct()
-                        logging.error(str(d))
-                        corrections.append(d)
-                except Exception as ex:
-                    if not self.skip_over_errors:
-                        raise
-                    else:
-                        corrections.append(
-                            {"errors": ["Bad handler " + str(h)],
-                             "actions": []})
-            return corrections
 
         for i, job in enumerate(self.jobs):
+            chk_fname = pjoin(cwd, "custodian.chk.{}.tar.gz".format(i))
+            if self.checkpoint and os.path.exists(chk_fname):
+                run_log = Custodian._load_checkpoint(chk_fname)
+                continue
             run_log.append({"job": job.to_dict, "corrections": []})
             for attempt in xrange(self.max_errors):
                 logging.info(
@@ -196,8 +214,9 @@ class Custodian(object):
                             if p.poll() is not None:
                                 break
                             if n % self.monitor_freq == 0:
-                                corrections = do_check(
-                                    self.monitors, terminate_func=p.terminate)
+                                corrections = _do_check(
+                                    self.monitors, terminate_func=p.terminate,
+                                    skip_over_errors=self.skip_over_errors)
                                 if len(corrections) > 0:
                                     has_error = True
                                     total_errors += len(corrections)
@@ -215,17 +234,18 @@ class Custodian(object):
                 else:
                     remaining_handlers = self.handlers
 
-                corrections = do_check(remaining_handlers)
+                corrections = _do_check(
+                    remaining_handlers,
+                    skip_over_errors=self.skip_over_errors)
                 if len(corrections) > 0:
                     has_error = True
                     total_errors += len(corrections)
                     run_log[-1]["corrections"].extend(corrections)
 
-                if self.log_file is not None:
-                    #Log the corrections to a json file.
-                    with open(self.log_file, "w") as f:
-                        logging.info("Logging to {}...".format(self.log_file))
-                        json.dump(run_log, f, indent=4)
+                #Log the corrections to a json file.
+                with open(Custodian.LOG_FILE, "w") as f:
+                    logging.info("Logging to {}...".format(Custodian.LOG_FILE))
+                    json.dump(run_log, f, indent=4)
 
                 # If there are no errors detected, perform postprocessing and
                 # exit.
@@ -243,12 +263,22 @@ class Custodian(object):
 
             if unrecoverable or total_errors >= self.max_errors:
                 break
+
+            # Checkpoint after each job so that we can recover from last
+            # point and remove old checkpoints
+            if self.checkpoint:
+                Custodian._save_checkpoint(cwd, i)
+
         end = datetime.datetime.now()
         logging.info("Run ended at {}.".format(end))
         run_time = end - start
         if total_errors >= self.max_errors:
             logging.info("Max {} errors reached. Exited..."
                          .format(self.max_errors))
+        elif not unrecoverable:
+            #Cleanup checkpoint files (if any) if run is successful.
+            Custodian._delete_checkpoints(cwd)
+
         logging.info("Run completed. Total time taken = {}.".format(run_time))
 
         if self.gzipped_output:
@@ -263,6 +293,26 @@ class Custodian(object):
         return run_log
 
 
+def _do_check(handlers, terminate_func=None, skip_over_errors=False):
+    corrections = []
+    for h in handlers:
+        try:
+            if h.check():
+                if terminate_func is not None and h.is_terminating:
+                    terminate_func()
+                d = h.correct()
+                logging.error(str(d))
+                corrections.append(d)
+        except Exception as ex:
+            if not skip_over_errors:
+                raise
+            else:
+                corrections.append(
+                    {"errors": ["Bad handler " + str(h)],
+                     "actions": []})
+    return corrections
+
+
 class ErrorHandler(object):
     """
     Abstract base class defining the interface for an ErrorHandler.
@@ -303,6 +353,15 @@ class ErrorHandler(object):
         """
         return False
 
+    @property
+    def is_terminating(self):
+        """
+        Whether this handler terminates a job upon error detection. Defaults
+        to True. Set to False for non-terminating errors, i.e., errors that
+        can be recovered from without terminating a job.
+        """
+        return True
+
     @abstractproperty
     def to_dict(self):
         """
@@ -392,12 +451,12 @@ def backup(filenames, prefix="error"):
             error.1.tar.gz, error.2.tar.gz, ... will be generated.
     """
     num = max([0] + [int(f.split(".")[1])
-                     for f in glob.glob("{}.*.tar.gz".format(prefix))])
+                     for f in glob("{}.*.tar.gz".format(prefix))])
     filename = "{}.{}.tar.gz".format(prefix, num + 1)
     logging.info("Backing up run to {}.".format(filename))
     tar = tarfile.open(filename, "w:gz")
     for fname in filenames:
-        for f in glob.glob(fname):
+        for f in glob(fname):
             tar.add(f)
     tar.close()
 
@@ -416,4 +475,4 @@ def gzip_dir(path):
             with open(f, 'rb') as f_in, \
                     GzipFile('{}.gz'.format(f), 'wb') as f_out:
                 f_out.writelines(f_in)
-            os.remove(f)
\ No newline at end of file
+            os.remove(f)
